/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// $Id: FifoDisk.cc 105 2006-11-10 02:52:34Z gregor $
//
#include <assert.h>
#include <unistd.h>
#include <pcap.h>
#include <string>

#include "config.h"
#ifdef HAVE_LIBPCAPNAV
#include <pcapnav.h>
#endif

#include "tm.h"
#include "FifoDisk.hh"
#include "Query.hh"


/***************************************************************************
 * FifoDisk
 */

FifoDisk::FifoDisk(const std::string& classname, u_fifosize_t size,
				   u_fifosize_t file_size, pcap_t* pcap_handle):
		classname(classname), size(size), file_size(file_size),
		tot_bytes(0), tot_pkts(0),
		file_number(0), pcap_handle(pcap_handle),
held_bytes(0), held_pkts(0), oldestTimestamp(0) {}


FifoDisk::~FifoDisk() {
	//	printf("~FifoDisk()\n");
	while (!files.empty()) {
		delete(files.back());
		files.pop_back();
	}
	//	printf("~FifoDisk() exit\n");
}


void FifoDisk::addPkt(const pkt_ptr p) {
	if (size>0) {
		if (files.empty() ||
				files.back()->getHeldBytes()
				+ sizeof(struct pcap_file_header)
				+ sizeof(struct pcap_pkthdr)
				+ ((struct pcap_pkthdr*)p)->caplen > file_size) {
			// need new file
			if (!files.empty()) {
				// close current file which just ran full
				files.back()->close();
				if (held_bytes+file_size > size) {
					// delete oldest file
					held_bytes-=files.front()->getHeldBytes();
					held_pkts-=files.front()->getHeldPkts();
					files.front()->remove();
					delete(files.front());
					files.pop_front();
					oldestTimestamp = files.front()->getOldestTimestamp();
				}
			}
			file_number++;
			const int strsz=classname.length()+1+8+1; // "_" + "number" + "\0"
			char *new_file_name=(char*)malloc(strsz);
			snprintf(new_file_name, strsz, "%s_%08x",
					 classname.c_str(), file_number);
			files.push_back(new FifoDiskFile(new_file_name, pcap_handle));
			free(new_file_name);
		}
		files.back()->addPkt(p);
		if (oldestTimestamp < 1e-3)
			oldestTimestamp = files.front()->getOldestTimestamp();
		held_bytes+=sizeof(struct pcap_pkthdr)+((struct pcap_pkthdr*)p)->caplen;
		held_pkts++;
		tot_bytes+=sizeof(struct pcap_pkthdr)+((struct pcap_pkthdr*)p)->caplen;
		tot_pkts++;
	}
}

tm_time_t FifoDisk::getOldestTimestamp() const {
	return oldestTimestamp;
}


pkt_count_t FifoDisk::query(QueryRequest *qreq, QueryResult *qres,
					 IntervalSet *interval_set) {
	pkt_count_t matches=0;
	IntervalSet::iterator i_i=interval_set->begin();
	std::list <FifoDiskFile*>::iterator f_i=files.begin();
	while ( f_i!=files.end() && i_i != interval_set->end() ) {
		matches += (*f_i)->query(interval_set, qreq, qres);
		f_i++;
	}
	return matches;
}


/***************************************************************************
 * FifoDiskFile
 */

FifoDiskFile::FifoDiskFile(const std::string& filename, pcap_t* pcap_handle):
		filename(filename), is_open(false), held_bytes(0), held_pkts(0),
		pcap_handle(pcap_handle),
		oldest_timestamp(0),
newest_timestamp(0) {
	//  pcap_handle=pcap_open_dead(0, 0);
	open();
}


void FifoDiskFile::open() {
	//	printf("opening file %s\n", filename.c_str());
	pcap_dumper_handle=pcap_dump_open(pcap_handle, filename.c_str());
	if (!pcap_dumper_handle) {
		char *pcap_errstr = pcap_geterr(pcap_handle);
		fprintf(stderr, "could not open file %s: %s\n",
				filename.c_str(), pcap_errstr);
	} else {
		is_open=true;
	}
}

FifoDiskFile::~FifoDiskFile() {
	if (is_open) close();
	//  pcap_close(pcap_handle);
}


void FifoDiskFile::remove() {
	if (is_open) close();
	unlink(filename.c_str());
}


void FifoDiskFile::close() {
	//  printf("closing file %s\n", filename.c_str());
	pcap_dump_close(pcap_dumper_handle);
	is_open=false;
}

void FifoDiskFile::addPkt(const struct pcap_pkthdr *header,
						  const unsigned char *packet) {
	assert(is_open==true);
	pcap_dump((u_char*)pcap_dumper_handle,
			  header,                         // pcap header
			  packet);                        // packet
	if (held_pkts==0) oldest_timestamp=to_tm_time(&header->ts);
	else newest_timestamp=to_tm_time(&header->ts);
	held_pkts++;
	held_bytes+=sizeof(struct pcap_pkthdr)+header->caplen;
}

void FifoDiskFile::addPkt(pkt_ptr p) {
	addPkt((struct pcap_pkthdr*)p,         // pcap header
		   p+sizeof(struct pcap_pkthdr));  // packet
}

pkt_count_t FifoDiskFile::query(IntervalSet *set, 
						 QueryRequest *qreq, QueryResult *qres) {
	pkt_count_t matches = 0;
	unsigned scanned_packets=0;
#ifdef HAVE_LIBPCAPNAV
	struct timeval tv1, tv2;
	struct timeval tv;
	int res;

	//fprintf(stderr, "FifoDiskFile::query\n");

	// FIXME: Protect the pcap_dumper_handle from capture thread!!
	if (is_open)
		flush();

	pcapnav_t *ph=pcapnav_open_offline(filename.c_str());
	if (!ph) {
		char *pcap_errstr = pcapnav_geterr(ph);
		fprintf(stderr, "FifoDiskFile::query: could not open file %s: %s\n",
				filename.c_str(), pcap_errstr);
	} else {
		struct pcap_pkthdr hdr;
		const u_char *pkt;
		for (IntervalSet::iterator it=set->begin(); it!=set->end(); it++) {
			assert(getNewestTimestamp() >= getOldestTimestamp());
			/* this should be handled by pcapnav_goto_timestamp.... 
			if (getOldestTimestamp() > (*it).getLast() ||
					getNewestTimestamp() < (*it).getStart() ) {
				fprintf(stderr, "Nicht im File: [%lf, %lf] <> [%lf,%lf]\n", 
						getOldestTimestamp(), getNewestTimestamp(), 
						(*it).getStart(), (*it).getLast());
				continue;
			}
			*/
			
			tv.tv_sec=(int)(*it).getStart();
			tv.tv_usec=(int)(1000000*((*it).getStart()-tv.tv_sec));
			
			if (pcapnav_get_timespan(ph, &tv1, &tv2) != 0) {
				printf("Could not obtain timespan.\n");
				  /* Rest of error handling */
			}
			// Check if interval overlaps trace start
			// FIXME: Don't hardcode the security margin with 1ms!!
			if ( (*it).getLast()+1e-3 >= to_tm_time(&tv1) &&
					(*it).getStart() <= to_tm_time(&tv1)) {
				res = PCAPNAV_DEFINITELY;
			}
			else 
				res = pcapnav_goto_timestamp(ph, &tv);
			switch(res) {
				case PCAPNAV_ERROR:
					fprintf(stderr, "pcapnav_goto_timestamp ERROR\n"); 
					log_file->log("query", "ERROR: pcapnav_goto_timestamp ERROR"); 
					break;
				case PCAPNAV_NONE:
				//	fprintf(stderr, "pcapnav_goto_timestamp NONE");
					break;
				case PCAPNAV_CLASH:
					fprintf(stderr, "ERROR: pcapnav_goto_timestamp CLASH\n"); 
					log_file->log("query", "ERROR: pcapnav_goto_timestamp CLASH"); 
					break;
				case PCAPNAV_PERHAPS:
					fprintf(stderr, "ERROR: pcapnav_goto_timestamp PERHAPS\n");
					log_file->log("query", "ERROR: pcapnav_goto_timestamp PERHAPS. interval=[%lf,%lf]. file=[%lf,%lf]",
							(*it).getStart(), (*it).getLast(), to_tm_time(&tv1), to_tm_time(&tv2));
					break;
				default:
					//fprintf(stderr, "DEFINITLY\n");
					break;
			}
			if (res != PCAPNAV_DEFINITELY) {
				continue;
			}
			do {
				pkt = pcapnav_next(ph, &hdr);
				scanned_packets++;
				if (!pkt)
					break;
				tm_time_t t=to_tm_time(&hdr.ts);
				if (t>(*it).getLast())
					break;
				if (qreq->matchPkt(&hdr, pkt))  {
					matches++;
					qres->sendPkt(&hdr, pkt);
				}
			} while (pkt);
		}
	}
	//DEBUG
	log_file->log("query", "FifoDiskFile::query [HAVE_LIBPCAPNAV] finished; matches %u; examined %u", 
			(unsigned)matches, scanned_packets);

	pcapnav_close(ph);
#else
	//TODO: Check Timespan!
	printf("FifoDiskFile::query file %s\n", getFilename().c_str());

	char errbuf[PCAP_ERRBUF_SIZE]="";
	pcap_t *ph=pcap_open_offline(filename.c_str(), errbuf);

	if (!ph) {
		char *pcap_errstr = pcap_geterr(ph);
		fprintf(stderr, "FifoDiskFile::query: could not open file %s: %s\n",
		        filename.c_str(), pcap_errstr);
	} else {
		int rc; /* pcap_next_ex rc meanings: 1 no problem, -2 no more packets */
		do {
			struct pcap_pkthdr *hdr;
			const u_char *pkt;
			rc=pcap_next_ex(ph, &hdr, &pkt);
			//printf("pcap_next_ex; rc=%d\n", rc);
			if (rc==1) {
				/*
				tm_time_t t=to_tm_time(&hdr->ts);
				while ( (*i)->getLast() < t && *i != set->end() ) *i++;
				if (t >= (*i)->getStart() && t <= (*i)->getLast()) {
				*/
				if (qreq->matchPkt(hdr, pkt)) {
					qres->sendPkt(hdr, pkt);
					matches++;
				}
				//	}
			}
		} while (rc==1);
	}
	//DEBUG
	fprintf(stderr, "FifoDiskFile::query [HAVE_LIBPCAPNAV] finished; matches %u\n", matches);
#endif
	return matches;
}
