/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// $Id: FifoMem.cc 104 2006-11-10 00:46:14Z gregor $
#include <pcap.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sched.h>

#include "tm.h"
#include "types.h"
#include "FifoMem.hh"
#include "Query.hh"
#include "Index.hh"


FifoMem::FifoMem(u_fifosize_t size): size(size) {
	//  printf("init FifoMem %u\n", size);
	start=(unsigned char *)malloc(size+MAXCAPLEN+sizeof(struct pcap_pkthdr)+1);
	assert(start);
	end=start+size;
	/*
	  unsigned char c=0;
	  for (unsigned char *i=buf_start; i<buf_start+size; i++)
	  *i=c++;
	  */

	// determine align positions
	align_gran=200;
	align_num=(size+MAXCAPLEN+sizeof(struct pcap_pkthdr))/align_gran + 1;
	align=(pkt_ptr*)malloc(align_num*sizeof(pkt_ptr));
	for (pkt_count_t i=0; i<align_num; i++) align[i]=start;
	//  printf ("FifoMem size=%"PRINTF_UINT64"; start=%p; end=%p; align len %d\n",size,start,end,align_num);
	s=wp=lp=start;
	a_wp=a_s=a_lp=a_max=0;
	a_next=start+align_gran;
	held_pkts=tot_pkts=tot_pktbytes=tot_lost_pkts=tot_lost_pktbytes=held_bytes=0;
	eviction_handler=NULL;

	pthread_mutex_init(&lock_mutex, NULL);
}


FifoMem::~FifoMem() {
//	printf("~FifoMem()\n");
	pthread_mutex_destroy(&lock_mutex);
	free(start);
	free(align);
//	printf("~FifoMem() exit\n");
}



void FifoMem::addPkt(const struct pcap_pkthdr *header,
					 const unsigned char *packet) {
	if (size>0) {

		lock();

		unsigned int pktlen=sizeof(struct pcap_pkthdr)+header->caplen;

		while (held_pkts && (wp<=s && s<wp+pktlen ||
							 wp<=s+size && s+size<wp+pktlen)) {
			// pktEviction eventually calls popPkt() which advances s
			if (pktEviction()) continue;
			assert(false);
		}

		memcpy(wp, header, sizeof(struct pcap_pkthdr));
		memcpy(wp+sizeof(struct pcap_pkthdr), packet, header->caplen);
		lp=wp;

		a_lp=a_wp;

		wp+=pktlen;

		//    if (a_next<=wp) a_lp=a_wp;

		while (a_next<=wp) {
			a_wp++;
			a_next+=align_gran;
			if (a_wp >= align_num) {
				fprintf(stderr, "a_wp=%d,  align_num=%d,  wp=%p,  start=%p, end=%p, size=%d,   s=%p\n", a_wp, align_num, wp, start, end, size, s);
				fprintf(stderr, "tot_pkts=%ld,  tot_pktbytes=%ld,  held_bytes=%ld\n", tot_pkts, tot_pktbytes, held_bytes);
			}
			assert(a_wp < align_num);
			if (a_next<wp) align[a_wp]=lp;
			else align[a_wp]=wp;
		}

		if (wp>=end) {
			wp-=size;
		}

		if (a_next>end) {
			a_next=start+align_gran;
			a_wp=0;
			align[0]=wp;
			while (a_next<=wp) {
				a_wp++;
				a_next+=align_gran;
				align[a_wp]=wp;
			}
		}

		tot_pkts++;
		tot_pktbytes+=header->caplen;
		held_pkts++;
		held_bytes+=header->caplen;

		/*
		printf("\nadded packet with length %d\n", pktlen);
		debugPrint();
		*/

		unlock();

	} // if (size>0)
}


pkt_ptr FifoMem::getWp() const {
	return wp;
}


pkt_ptr FifoMem::getS() const {
	return s;
}


pkt_count_t FifoMem::getTotPkts() const {
	return tot_pkts;
}


byte_count_t FifoMem::getTotPktbytes() const {
	return tot_pktbytes;
}


pkt_count_t FifoMem::getTotLostPkts() const {
	return tot_lost_pkts;
}


byte_count_t FifoMem::getTotLostPktbytes() const {
	return tot_lost_pktbytes;
}


u_fifosize_t FifoMem::popPkt() {
	u_fifosize_t n=0;
	if (held_pkts) {
		n=((struct pcap_pkthdr *)s)->caplen;
		s+=sizeof(struct pcap_pkthdr)+((struct pcap_pkthdr *)s)->caplen;
		if (s>=end) {
			s-=size;
			a_s=0;
		}
		while (align[a_s]<s) {
			a_s++;
		}
		held_pkts--;
		held_bytes-=n;
	}
	return n;
}

byte_count_t FifoMem::pktEviction() {
	if (eviction_handler)
		return eviction_handler->pktEviction();
	else
		return 0;
}


void FifoMem::setEvictionHandler(FifoMemEvictionHandler *h) {
	eviction_handler=h;
}


tm_time_t FifoMem::getOldestTimestamp() const {
	return to_tm_time(&((struct pcap_pkthdr*)s)->ts);
}


void FifoMem::debugPrint() const {
	debugPrint(stderr);
}
	
void FifoMem::debugPrint(FILE *fp) const {
	fprintf(fp, "\nstart = %ld  s = %ld  wp = %ld  lp = %ld  end = %ld  held_pkts = %lu\n",
		   start-start, s-start, wp-start, lp-start, end-start, held_pkts);
	fprintf(fp, "a_next = %ld  a_wp = %lu  a_s = %lu  a_lp = %lu  a_max = %lu\n",
		   a_next-start, a_wp, a_s, a_lp, a_max);

	fprintf(fp, "align={");
	for (pkt_count_t i=0; i<(align_num>10 ? 10 : align_num); i++)
		fprintf(fp, i ? ", %ld" : "%ld", align[i]-start);
	fprintf(fp, "}\n");

	/*
	align[i] ?
	to_tm_time(&((pcap_pkthdr*)align[i])->ts) :
	0
	*/

	/*
	#define ARRLEN 4
	#define WIDTH 70
	const pkt_ptr *ptrs[] = { &start, &s, &wp, &end };
	const char *names[] = { "start", "s", "wp", "end" };
	char out[ARRLEN][WIDTH+1];
	for (int i=0; i<ARRLEN; i++) {
	  memset(out[i], ' ', WIDTH); out[i][WIDTH]='\0';
	}
	for (int i=0; i<ARRLEN; i++) {
	  int pos = (int)( (double)(*ptrs[i]-start)/(end-start)*(WIDTH-1) );
	  if (pos < 0 || pos >= WIDTH) continue;
	  out[0][pos] = '|';
	  if (pos+strlen(names[i]) > WIDTH) pos -= pos+strlen(names[i]) - WIDTH;
	  int j;
	  for (j=1;
	j<ARRLEN && strspn(out[j]+pos, " ") >= strlen(names[i]);
	j++);
	  sprintf(out[j]+pos, names[i]);
	}
	for (int i=0; i<ARRLEN; i++) printf("%s\n", out[i]);
	*/
}

inline tm_time_t pkt_t (pkt_ptr p) {
	return to_tm_time(&((struct pcap_pkthdr*)p)->ts);
}

inline pkt_ptr FifoMem::block (pkt_ptr p) {
	assert (p<end+size);
	return p<end ? p : p-size;
}

bool FifoMem::bin_search (pkt_ptr *p, tm_time_t t, bool floor) {
	pkt_count_t a0=a_s, a1=a_lp, at;
	pkt_ptr pt;
	if (a1<a0) a1+=align_num;

	if ( t < pkt_t(align[a0<align_num ? a0 : a0-align_num] )
			//       && t >= pkt_t(s)
	   ) {
		*p=s;
		return true;
	} else if ( t > pkt_t(align[a1<align_num ? a1 : a1-align_num] )
				//	      && t <= pkt_t(lp)
			  ) {
		*p=lp;
		return true;
	};

	do {
		at=(a0+a1)/2;
		pt=align[at<align_num ? at : at-align_num];
		tm_time_t tt=pkt_t(pt);
		if ( pkt_t(align[a0<align_num ? a0 : a0-align_num]) == t ) {
			*p=align[a0<align_num ? a0 : a0-align_num];
			return true;
		} else if ( pkt_t(align[a1<align_num ? a1 : a1-align_num]) == t ) {
			*p=align[a1<align_num ? a1 : a1-align_num];
			return true;
		} else if ( tt > t ) a1=at;
		else /* tt < t */ a0=at;
	} while (a0<a1-1);

	if (floor) *p = align[a0<align_num ? a0 : a0-align_num];
	else *p=align[a1<align_num ? a1 : a1-align_num];

	return true;

}
//FIXME: When the interval is larger than the timepsan of packets
//in the FifoMem, we will get problems!!!
pkt_count_t FifoMem::query(QueryRequest *qreq, QueryResult *qres,
					IntervalSet *interval_set) {
	pkt_ptr p, end_ptr;
	pkt_count_t matches = 0;

	lock();


	for (IntervalSet::iterator i=interval_set->begin();
			i!=interval_set->end(); i++) {
		bool found;
		if (! (found=bin_search(&p, i->getStart(), true)))
			printf("%f not found\n", i->getStart());

		if (found) {
			if (p<wp)  
				end_ptr = wp;
			else 
				end_ptr = wp+size;

			while ((p<end_ptr) && pkt_t(block(p)) <= i->getLast() ) {
				if (qreq->matchPkt(block(p)))  {
					qres->sendPkt(block(p));
					matches++;
				}
				p+=sizeof(struct pcap_pkthdr)+
					((struct pcap_pkthdr*)block(p))->caplen;
			}
		}
	}
	//DEBUG:
	//fprintf(stderr, "FifoMem::query had %lu maches\n", matches);

	unlock();
	return matches;
}


void FifoMem::lock() {
	pthread_mutex_lock(&lock_mutex);
}

void FifoMem::unlock() {
	pthread_mutex_unlock(&lock_mutex);
}

