/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/* Storage.cc
 * $Id: Storage.cc 104 2006-11-10 00:46:14Z gregor $
 *
 */

#include <pcap.h>
#include <pthread.h>

#include <sstream>
#include <iostream>

#include "types.h"
#include "Storage.hh"
#include "Index.hh"
#include "IndexField.hh"
#include "FifoDisk.hh"
#include "packet_headers.h"
#include "conf.h"

#define SNAPLEN 8192


/***************************************************************************
 * callback handler for pcap_loop
 */

void callback(u_char *args, const struct pcap_pkthdr *header,
			  const u_char *packet) {
	Storage *storage = (Storage *)args;

	//FIXME: put these counters somewhere else
	tot_pkt_cnt++;
	tot_bytes+=header->len;

	/* FIXME: make this useful again 
	if (timercmp(&header->ts, &network_time, <)) {
		struct timeval dt;
		timersub(&header->ts, &network_time, &dt);
		log_file->log("main", "network time jumped backward %fs",-to_tm_time(&dt));
	}
	network_time=header->ts;
	*/ 

	storage->addPkt(header, packet);

} // callback()

/***************************************************************************
 * capture thread
 */

void *capture_thread(void *arg) {
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	Storage *storage = (Storage *)arg;

	//  sleep(15);
	pcap_loop(storage->ph, -1, (pcap_handler)callback, (u_char*)storage);
	return NULL;
}


Storage::Storage(StorageConfig& conf):
		snaplen(SNAPLEN),
		indexes(new Indexes),
		conns(250000),
tot_num_queries(0) {
	//FIXME: Deallocate fifos when throwing an exception !!!
	//FIXME: Same for indexes
	char errbuf[PCAP_ERRBUF_SIZE]; 
	struct bpf_program fp;

	conn_timeout = conf.conn_timeout;
	conns.setMaxSubscriptions(conf.max_subscriptions);

	// Get pcap handle 
	errbuf[0] = '\0';
	if (!conf.readtracefile.empty()) {
		ph = pcap_open_offline(conf.readtracefile.c_str(), errbuf);
	}
	else if (!conf.device.empty()) {
		ph = pcap_open_live(conf.device.c_str(), snaplen, 1, 20, errbuf);
	}
	else {
		throw std::string("either give an interface or a pcap trace");
	}

	if (!ph) {
		throw std::string("pcap error: ") + errbuf;
	}

	//TODO: If strlen(errbuf)>0, then errbuf contains a WARNING!
	//
	
	if (!conf.filter.empty()) {
		char *filterstr = strdup(conf.filter.c_str());
		if(pcap_compile(ph, &fp, filterstr, 0, 0)  < 0) {
			std::string errstring;
			errstring.append("Could not set capture filter. ");
			errstring.append("pcap_compile error: ");
			errstring.append(pcap_geterr(ph));
			pcap_close(ph);
			free(filterstr);
			throw errstring;
		}
		free(filterstr);

		if(pcap_setfilter(ph, &fp) < 0) {
			std::string errstring;
			errstring.append("pcap_setfilter error: ");
			errstring.append(pcap_geterr(ph));
			pcap_freecode(&fp);
			pcap_close(ph);
			throw errstring;
		}
		pcap_freecode(&fp);
	}
	for (std::list<Fifo*>::iterator it=conf.fifos.begin(); it!=conf.fifos.end(); it++) {
		(*it)->setPcapHandle(ph);
		fifos.push_back(*it);
	}
	conf.fifos.clear();

/* INDEX CONFIGRATION */
	/* comment any of the following blocks if you want to disable the 
	 * index */
	IndexType *idx = new Index<ConnectionIF4>(30, 1.0, int(250000), this);
	pthread_create(&(idx->maintainer_thread), NULL, start_index_thread, idx);
	indexes->addIndex(idx);
	
	idx = new Index<ConnectionIF3>(30, 3.0, int(250000), this);
	pthread_create(&(idx->maintainer_thread), NULL, start_index_thread, idx);
	indexes->addIndex(idx);

	idx = new Index<ConnectionIF2>(30, 5.0, int(250000), this);
	pthread_create(&(idx->maintainer_thread), NULL, start_index_thread, idx);
	indexes->addIndex(idx);

	idx = new Index<IPAddress>(30, 0, int(250000), this);
	pthread_create(&(idx->maintainer_thread), NULL, start_index_thread, idx);
	indexes->addIndex(idx);
/* END INDEX CONFIGRATION */


	for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
		(*i)->start();

	// Start Capture thread
	int i=pthread_create(&capture_thread_tid, NULL, capture_thread, (void *)this);
	if (i!=0) {
		pcap_close(ph);
		throw std::string("Could not create capture thread.");
	}
}

Storage::~Storage() {
	printf("~Storage\n");
	/*fprintf(stderr, "Breaking pcap_loop()\n");
	pcap_breakloop(ph);
	fprintf(stderr, "pcap_loop() is destroyed\n"); */
	for (std::list<Fifo*>::iterator it=fifos.begin(); it!=fifos.end(); it++)
		delete (*it);
	fprintf(stderr, "fifos deleted\n");
	delete indexes;
	fprintf(stderr, "pcap handle closed\n");
	pcap_close(ph);
}

void Storage::cancelThread() {
	fprintf(stderr, "Canceling capture thread\n");
	pthread_cancel(capture_thread_tid);
	fprintf(stderr, "Joining capture thread\n");
	pthread_join(capture_thread_tid, NULL);
	fprintf(stderr, "Capture thread is gone.\n");
	indexes->cancelThread();
}


/*void Storage::addFifo(Fifo* f) {
	fifos.push_back(f);
}*/


void Storage::addPkt(const struct pcap_pkthdr *header,
					 const unsigned char *packet) {
	u_int16_t ether_type=ntohs(ETHERNET(packet)->ether_type);
	if ( ! (ether_type==ETHERTYPE_IP || ether_type==0x8100) ) {
		//    fprintf(stderr,"unknown ether_type 0x%.4X\n", ether_type);
		return;
	}
	const unsigned char* idxpacket=packet;
	// skip VLAN header for indexing
	if (ether_type==0x8100) idxpacket+=4;

	tm_time_t now = to_tm_time(&header->ts);

	/* update connections state, classify, elephant cutoff */
	conns.lock();
	conns.removeOld(now-conn_timeout);
	Connection* c=conns.addPkt_get(header, idxpacket);
	Fifo *f=c->getFifo();
	QueryResult *qr=c->getSubscription();
	conns.unlock();
	if (!f) {
		int max_precedence=INT_MAX;
		for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
			if (// packet matches this class' filter and
				(*i)->matchPkt(header, packet) &&
				// first match or higher precedence match
				(f==NULL || (*i)->getPrecedence()>max_precedence) ) {
				f=*i;
				max_precedence=f->getPrecedence();
			}
	} // if (!f)
	if (f) {
		c->setFifo(f);
		bool tcp_ctrl_flag=false;
		if (IP(packet)->ip_p==IPPROTO_TCP)
			if (TCP(packet)->th_flags & ( TH_FIN | TH_SYN | TH_RST ))
				tcp_ctrl_flag=true;
		if ( ( c->getSuspendCutoff() | tcp_ctrl_flag )
				&& f->addPkt(header, packet, NULL) ||
				f->addPkt(header, packet, c)) {
			/* packet was stored in Fifo *f */
			/* update indexes */
			for (std::list<IndexType*>::iterator i=indexes->begin();
					i!=indexes->end();
					i++) {
				(*i)->addPkt(header, idxpacket);
				//if (conf_main_max_index_entries &&
				//		(*i)->getNumEntriesRAM() > conf_main_max_index_entries)
				//	(*i)->removeCount(10);
			}
		} // if (f->addPkt) ... else it was "cut off"
	} // if (f)
	if (qr) {
		// there is a subscription for this connection
		qr->sendPkt(header, packet);
	}
}


void Storage::aggregateIndexFiles() {
	for (std::list<IndexType*>::iterator i=indexes->begin(); i!=indexes->end(); i++) {
		(*i)->aggregate();
	}
}

void Storage::debugPrint() {
	debugPrint(stderr);
}

void Storage::debugPrint(FILE *fp) {
	//  Fifo* f[]={&f_udp, &f_tcp, &f_other};
	for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
		fprintf(fp, "%s totBytes: %"PRINTF_UINT64"\n", (*i)->getClassname().c_str(),
			   (*i)->getFm()->getTotPktbytes());
	for (std::list<IndexType*>::iterator i=indexes->begin();
			i!=indexes->end();
			i++) {
		//(*i)->lock();
		fprintf(fp, "# %s index nodes RAM/Disk %"PRINTF_UINT64" %"PRINTF_UINT64"\n",
			   (*i)->getIndexName().c_str(),
			   (*i)->getNumEntriesRAM(),
			   (*i)->getNumEntriesDisk());
		(*i)->debugPrint(fp);
		//(*i)->unlock();
	}
}


void Storage::logStatsClasses() {
	for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
		log_file->log((*i)->getClassname(), (*i)->getStatsStr());
}


std::string Storage::getStatsIndexesStr() {
	std::stringstream ss;
	for (std::list<IndexType*>::iterator i=indexes->begin();
			i!=indexes->end();
			i++) {
		//(*i)->lock();
		ss << (*i)->getNumEntriesRAM() << " RAM / "
		<< (*i)->getNumEntriesDisk() << " Disk "
		<< (*i)->getIndexName() << " index nodes  ";
		//    (*i)->DBStatPrint();
		//(*i)->unlock();
	}
	return ss.str();
}


tm_time_t Storage::getOldestTimestampMem() {
	tm_time_t r=0;
	for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
		if (r==0 || (*i)->getFm()->getOldestTimestamp() < r)
			r=(*i)->getFm()->getOldestTimestamp();
	return r;
}


tm_time_t Storage::getOldestTimestampDisk() {
	tm_time_t r=0;
	for (std::list<Fifo*>::iterator i=fifos.begin(); i!=fifos.end(); i++)
		if (r==0 || (*i)->getFd()->getOldestTimestamp() < r)
			r=(*i)->getFd()->getOldestTimestamp();
	return r;
}


void Storage::query(QueryRequest *query_req, QueryResult *query_res) {
	struct timeval t_start, t_end;
	gettimeofday(&t_start, NULL);

	/*
	printf("Storage::query %s\n", query_req->getField()->getStr().c_str());
	printf("  IndexType* idx=indexes.getIndexByName(%s)\n", 
	query_req->getField()->getIndexName().c_str());
	*/

	IndexType* idx=indexes->getIndexByName(query_req->getField()->getIndexName());
	if (!idx) {
		printf("no index for %s\n",
			   query_req->getField()->getIndexName().c_str());
	} else {
		unsigned matches=0;
		/*
		fprintf(stderr, "have index for %s\n",
		 query_req->getField()->getIndexName().c_str());
		fprintf(stderr, "query key is \"%s\"\n", query_req->getField()->getStr().c_str());
		*/
		IntervalSet interval_set;
		/* NOTE: it may happen, that the index is rotated between the
		 * lookupMem and lookupDisk call. This isn't a problem. In the worst case
		 * an entry will be found twice: in the memory index before the rotation
		 * and in the disk index  after the rotation. Since the Interval sets
		 * ensures sort order and overlap free intervals, we are fine. No need
		 * for locks or something. 
		 */
		
		//    printf("getting ie for %s\n", query_req->getField()->getStr().c_str());
		idx->lookupMem(&interval_set, query_req->getField());

		if (!query_req->isMemOnly()) {
			idx->lookupDisk(&interval_set, query_req->getField(),
							query_req->getT0(), query_req->getT1());
		}
		//XXX:
		log_file->log("query", "interval_set now has %u entries", interval_set.getNumIntervals());

		//    printf("disk_set: %s'n", disk_set->getStr().c_str());

		for (std::list<Fifo*>::iterator fifo_i=fifos.begin();
				fifo_i != fifos.end(); fifo_i++)
			matches+=(*fifo_i)->query(query_req, query_res, &interval_set);
		log_file->log("query", "Query is done. Had %u matches", matches);

	} // if (!idx)
	/*  printf("idx query fin\n"); */

//TODO:
#if 0
	/* handle potential subscription request */
	if (query_req->getT1() == 0 || query_req->getT1() == -1) {
		/* subscription requested */
		if (query_req->getField()->getIndexName() == "connection4") {
			/* it's a request for a connection */
			if (conns.getMaxSubscriptions()
					&& conns.getSubscriptions() >= conns.getMaxSubscriptions()) {
				/* maximum number of subscriptions active */
				printf("maximum number of subscriptions is active, "
					   "dropping request\n");
			} else {
				if (query_req->getT1() == -1) {
					/* timeout suspension requested, too */
					suspendTimeout(/* we know for sure this is a ConnectionIF */
						*( (ConnectionIF4*)
						   (query_req->getField()) ) -> getCID(), true  );
				}
				conns.lock();
				Connection *c=conns.get(
								  /* we know for sure this is a ConnectionIF */
								  ( (ConnectionIF4*)
									(query_req->getField()) ) -> getCID()  );
				if (c) {
					/* connection is in connection table */
					c->setSubscription(query_res);
					conns.increaseSubscriptions();
					query_res=NULL; /* don't delete it at end of this function */
				} else {
					/* connection is NOT in connection table */
					//	  fprintf(stderr, "cannot subscribe to connection: not in connection table: %s\n", query_req->getField()->getStr().c_str());
				}
				conns.unlock();
			} /* if (max_subscriptions) ... */
		} else {
			/* it's a request for something different */
			printf("cannot subscribe to request for another thing than a connection\n");
		}
	} /* if (subscription requested) */
#endif

	if (query_res) {
		/* haven't passed it on to a subscription, delete it */
		delete query_res;
	}
	delete query_req;

	gettimeofday(&t_end, NULL);

	tot_queries_duration+=(u_int64_t) ( (t_end.tv_sec-t_start.tv_sec)*1e6
										+(t_end.tv_usec-t_start.tv_usec) );
	tot_num_queries++;
}

bool Storage::suspendCutoff(ConnectionID4 cid, bool b) {
	bool rv=false;
	conns.lock();
	Connection *c=conns.get(&cid);
	if (c) {
		c->setSuspendCutoff(b);
		rv=true;
	} else {
		//    fprintf(stderr, "Storage::suspend_cutoff(%d): CONNECTION NOT IN CONNECTION TABLE\n", b);
		rv=false;
	}
	conns.unlock();
	return rv;
}

bool Storage::suspendTimeout(ConnectionID4 cid, bool b) {
	bool rv=false;
	conns.lock();
	Connection *c=conns.get(&cid);
	if (c) {
		c->setSuspendTimeout(b);
		rv=true;
	} else {
		//    fprintf(stderr, "Storage::suspend_timeout(%d): CONNECTION NOT IN CONNECTION TABLE\n", b);
		//XXX: c=conns.newConn_get(new ConnectionID4(cid));
		//c->setSuspendTimeout(b);
		rv=false;
	}
	conns.unlock();
	return rv;
}

Fifo* Storage::getFifoByName(std::string search_name) {
	Fifo *r=NULL;
	for (std::list<Fifo*>::iterator i=fifos.begin();
			!r && i!=fifos.end();
			i++)
		if ((*i)->getClassname() == search_name) r=*i;
	return r;
}

